﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Services.Client;

using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;

using mana.Data.Azure.Tables;
using mana.Data.Azure.Queues;

namespace Library
{
    [Serializable]
    public class Contact
    {
        public Contact(string contactId, string name, DateTime creationDate)
        {
            ContactId = contactId;
            Name = name;
            CreationDate = creationDate;

            _emails = new List<EmailAddress>();
            _phones = new List<PhoneNumber>();
        }

        #region Properties

        public string ContactId { get; private set; }
        public string Name { get; set; }
        public DateTime CreationDate { get; private set; }

        private bool _phonesLoaded;
        private List<PhoneNumber> _phones;
        public IEnumerable<PhoneNumber> Phones 
        {
            get 
            {
                if (!_phonesLoaded)
                    LoadPhones();

                return _phones; 
            }
        }

        private bool _emailsLoaded;
        private List<EmailAddress> _emails;
        public IEnumerable<EmailAddress> Emails
        {
            get 
            {
                if (!_emailsLoaded)
                    LoadEmails();

                return _emails; 
            }
        }

        private ContactReport _report;
        public ContactReport Report
        {
            get 
            {
                if (_report == null)
                    _report = new ContactReport(ContactId);

                return _report;
            }
        }

        #endregion

        #region Methods

        #region Instance methods

        public void LoadPhones()
        {
            try
            {
                if (string.IsNullOrWhiteSpace(ContactId))
                    throw new Exception("No contact id specified.");

                _phones = PhoneNumber.GetByContact(ContactId);
                _phonesLoaded = true;
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's phones couldn't be loaded. " + ex.Message);
            }
        }

        public void LoadEmails()
        {
            try
            {
                if (string.IsNullOrWhiteSpace(ContactId))
                    throw new Exception("No contact id specified.");

                _emails = EmailAddress.GetByContact(ContactId);
                _emailsLoaded = true;
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's phones couldn't be loaded. " + ex.Message);
            }
        }

        public void Update()
        {
            try
            {
                DataAccess.Update(this);
            }
            catch (Exception ex)
            {
                throw new Exception("The contact couldn't be updated. " + ex.Message);
            }
        }

        public PhoneNumber AddPhone(string number, PhoneNumberType type)
        {
            try
            {
                var phone = PhoneNumber.Add(ContactId, number, type);
                _phones.Add(phone);

                return phone;
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's phone couldn't be added. " + ex.Message);
            }
        }

        public IEnumerable<PhoneNumber> AddPhones(IEnumerable<string> numbers, PhoneNumberType type)
        {
            try
            {
                var phones = PhoneNumber.Add(ContactId, numbers, type);
                _phones.AddRange(phones);

                return phones;
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's phones couldn't be added. " + ex.Message);
            }
        }

        public void DeletePhone(PhoneNumber phone)
        {
            try
            {
                PhoneNumber.Delete(phone);
                _phones.Remove(phone);
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's phone couldn't be deleted. " + ex.Message);
            }
        }

        public void ClearPhones()
        {
            try
            {
                PhoneNumber.DeleteByContact(ContactId);
                _phones.Clear();
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's phone couldn't be cleared. " + ex.Message);
            }
        }

        public EmailAddress AddEmail(string emailAddress, EmailAddressType type)
        {
            try
            {
                var email = EmailAddress.Add(ContactId, emailAddress, type);
                _emails.Add(email);

                return email;
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's email couldn't be added. " + ex.Message);
            }
        }

        public IEnumerable<EmailAddress> AddEmails(IEnumerable<string> addresses, EmailAddressType type)
        {
            try
            {
                var emails = EmailAddress.Add(ContactId, addresses, type);
                _emails.AddRange(emails);

                return emails;
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's emails couldn't be added. " + ex.Message);
            }
        }

        public void DeleteEmail(EmailAddress email)
        {
            try
            {
                EmailAddress.Delete(email);
                _emails.Add(email);
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's emails couldn't be deleted. " + ex.Message);
            }
        }

        public void ClearEmails()
        {
            try
            {
                EmailAddress.DeleteByContact(ContactId);
                _emails.Clear();
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's emails couldn't be cleared. " + ex.Message);
            }
        }

        #endregion 

        #region Static methods

        public static Contact Add(string name)
        {
            string contactId = Guid.NewGuid().ToString().Replace("-", string.Empty);
            var contact = new Contact(contactId, name, DateTime.Now);

            DataAccess.Add(contact);

            return contact;
        }

        public static void Delete(Contact contact)
        {
            DataAccess.Delete(contact);

            contact.ClearEmails();
            contact.ClearPhones();
        }

        public static Contact Get(string contactId)
        {
            return DataAccess.Get(contactId);
        }

        public static List<Contact> GetByName(string name)
        {
            return DataAccess.GetByName(name);
        }

        public static List<Contact> GetAll()
        {
            return DataAccess.GetAll();
        }

        #endregion

        #endregion

        #region Internal classes

        //  I create this class as "private" to avoid other classes to manipulate Contact's data directly 
        private static class DataAccess
        {
            #region Methods

            public static void Add(Contact contact)
            {
                var entity = new ContactEntity
                {
                    ContactId = contact.ContactId,
                    Name = contact.Name,
                    CreationDate = contact.CreationDate
                };

                var ctx = new ContactsContext();
                ctx.Add(entity);
            }

            public static void Delete(Contact contact)
            {
                if (contact == null)
                    return;

                string partitionKey = ContactEntity.GeneratePartitionKey();
                string rowKey = ContactEntity.GenerateRowKey(contact.ContactId);

                var ctx = new ContactsContext();
                ctx.Delete(partitionKey, rowKey);
            }

            public static void Update(Contact contact)
            {
                if (contact == null)
                    throw new ArgumentNullException("contact");

                string partitionKey = ContactEntity.GeneratePartitionKey();
                string rowKey = ContactEntity.GenerateRowKey(contact.ContactId);

                var ctx = new ContactsContext();
                
                var entity = ctx.Get(partitionKey, rowKey);
                if (entity == null)
                    throw new Exception(string.Format("The contact '{0}' to update doesn't exist.", contact.ContactId));

                //  The only property that we'll allow to update is the name
                //  So that's the only entity's property we'll set
                entity.Name = contact.Name;

                ctx.Update(entity);
            }

            public static Contact Get(string contactId)
            {
                string partitionKey = ContactEntity.GeneratePartitionKey();
                string rowKey = ContactEntity.GenerateRowKey(contactId);

                var ctx = new ContactsContext();
                var entity = ctx.Get(partitionKey, rowKey);
                return MaterializeFromEntity(entity);
            }

            //  There can be more than 1 contact with the same name
            //  That's the reason why we're returning a list
            public static List<Contact> GetByName(string name)
            {
                var ctx = new ContactsContext();

                var query =
                    from    entity in ctx.Entities
                    where   entity.PartitionKey == ContactEntity.GeneratePartitionKey() && entity.Name == name
                    select  entity;

                var entities = ctx.Get(query);
                return MaterializeCollectionFromEntities(entities);
            }

            public static List<Contact> GetAll()
            {
                var ctx = new ContactsContext();
                var entities = ctx.GetAll();
                return MaterializeCollectionFromEntities(entities);
            }

            private static Contact MaterializeFromEntity(ContactEntity entity)
            {
                if (entity == null)
                    return null;

                return new Contact(entity.ContactId, entity.Name, entity.CreationDate);
            }

            private static List<Contact> MaterializeCollectionFromEntities(IEnumerable<ContactEntity> entities)
            {
                var list = new List<Contact>();
                foreach (var entity in entities)
                {
                    var obj = MaterializeFromEntity(entity);
                    if (obj != null)
                        list.Add(obj);
                }
                return list;
            }

            #endregion

            #region Internal classes

            class ContactEntity : TableServiceEntity
            {
                public ContactEntity()
                {
                    PartitionKey = GeneratePartitionKey();
                }

                #region Properties

                private string _contactId;
                public string ContactId 
                {
                    get { return _contactId; }
                    set
                    {
                        _contactId = value;
                        RowKey = GenerateRowKey(value);
                    }
                }

                public string Name { get; set; }
                public DateTime CreationDate { get; set; }

                #endregion

                #region Methods

                public static string GeneratePartitionKey()
                {
                    return "contacts";
                }

                public static string GenerateRowKey(string contactId)
                {
                    return contactId;
                }

                #endregion
            }

            class ContactsContext : GenericTableContext<ContactEntity>
            {
                /// <summary>
                /// Initializes context using a parameterized Cloud Storage Account.
                /// </summary>
                public ContactsContext()
                    : this(Parameters.Internal.GetCloudStorageAccount())
                { }

                /// <summary>
                /// Initializes context using the specified Cloud Storage Account.
                /// </summary>
                /// <param name="account"></param>
                public ContactsContext(CloudStorageAccount account)
                    : base(account, "Contacts")
                { }
            }

            #endregion
        }

        #endregion
    }

    public static class ContactListExtensions
    {
        public static Contact FindById(this IEnumerable<Contact> list, string contactId)
        {
            return list.FirstOrDefault(c => c.ContactId == contactId);
        }
    }

    [Serializable]
    public class ContactReport
    {
        public ContactReport(string contactId)
        {
            ContactId = contactId;
            Filename = string.Format("Contact Report - {0} ({1}).txt", contactId, DateTime.Now.ToString("yyyy-MM-dd HH:mm"));
        }

        #region Properties

        public string ContactId { get; private set; }
        public string Filename { get; private set; }
        
        private CloudBlob _blob;
        private CloudBlob Blob
        {
            get 
            {
                if (_blob == null)
                    _blob = Parameters.Internal.GetContactsReportsContainer().GetBlobReference(Filename);

                return _blob;
            }
        }

        #endregion

        public void EnqueueGeneration()
        {
            var queue = new ContactReportGenerationQueue();
            queue.Enqueue(this);
        }

        public void Generate()
        {
            //  inside here we'll define the generation logic for the contact specified by ContactId
            try
            {
                var contact = Contact.Get(ContactId);
                if (contact == null)
                    throw new Exception(string.Format("The contact '{0}' doesn't exist.", ContactId));

                byte[] bytesContent = GenerateReportContent(contact);
                Blob.UploadByteArray(bytesContent);
            }
            catch (Exception ex)
            {
                throw new Exception(string.Format("The contact '{0}' report couldn't be generated. {1}", ContactId, ex.Message));
            }
        }

        private byte[] GenerateReportContent(Contact contact)
        {
            if (contact == null)
                throw new ArgumentNullException("contact");

            contact.LoadEmails();
            contact.LoadPhones();

            var sb = new StringBuilder();
            sb.AppendLine(string.Format("Contact: {0} ({1})", contact.Name, contact.ContactId));
            sb.AppendLine();

            int phoneNumbersCount = contact.Phones.Count();
            sb.AppendLine(string.Format("Phone numbers ({0}):", phoneNumbersCount));
            if (phoneNumbersCount > 0)
            {
                foreach (var phoneNumber in contact.Phones)
                {
                    sb.AppendLine(string.Format("{0} ({1}", phoneNumber.Number, phoneNumber.Type));
                }
            }
            else
            {
                sb.AppendLine("No phone numbers.");
            }
            sb.AppendLine();

            int emailAddressesCount = contact.Emails.Count();
            sb.AppendLine(string.Format("Phone numbers ({0}):", emailAddressesCount));
            if (emailAddressesCount > 0)
            {
                foreach (var emailAddress in contact.Emails)
                {
                    sb.AppendLine(string.Format("{0} ({1}", emailAddress.Address, emailAddress.Type));
                }
            }
            else
            {
                sb.AppendLine("No email addresses.");
            }

            return Encoding.UTF8.GetBytes(sb.ToString());
        }
    }

    public class ContactReportGenerationQueue : GenericQueue<ContactReport>
    {
        public ContactReportGenerationQueue()
            : base(Parameters.Internal.GetCloudStorageAccount(), "contact-reports-generation")
        {
            this.PoisonMessageFound += ContactReportGenerationQueue_PoisonMessageFound;

            this.Processor.Processing += Processor_Processing;
            this.Processor.Error += Processor_Error;
        }

        void Processor_Error(object sender, QueueProcessorErrorEventArgs<ContactReport> e)
        {
            //  This event is raised whenever an error ocurred during the processing
            //  In here you can add some logic to store the error and the item that was being 
            //  processed during the error using the properties "e.Exception" and "e.Item"
        }

        void Processor_Processing(object sender, QueueProcessorProcessingEventArgs<ContactReport> e)
        {
            //  This event is raised to process after an object is dequeud from the queue
            //  After this event's successful execution the dequeued object will be deleted from the queue

            var contactReport = e.Item;
            contactReport.Generate();
        }

        void ContactReportGenerationQueue_PoisonMessageFound(object sender, GenericQueuePoisonMessageEventArgs e)
        {
            //  Whenever a poison message is found this event is raised
            //  In here you can add some logic to store the poison message for debugging purposes
        }

        protected override string GetMessageFromObj(ContactReport obj)
        {
            return obj.ContactId;
        }

        protected override ContactReport GetObjFromMessage(CloudQueueMessage message)
        {
            string contactId = message.AsString;
            var contact = Contact.Get(contactId);
            if (contact == null)
                return null;

            return contact.Report;
        }

        protected override bool IsPoisonMessage(CloudQueueMessage message)
        {
            //  What makes a message poison?
            //  If there's not a contactId in the message, then it's a poison message
            string contactId = message.AsString;
            return string.IsNullOrWhiteSpace(contactId);
        }
    }
}
