﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Services.Client;

using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;

using mana.Data.Azure.Tables;

namespace Library
{
    public enum EmailAddressType
    {
        Personal = 0,
        Work = 1
    }

    [Serializable]
    public class EmailAddress
    {
        public EmailAddress(string contactId, string emailAddressId, string address, EmailAddressType type)
        {
            ContactId = contactId;
            EmailAddressId = emailAddressId;
            Address = address;
            Type = type;
        }

        public string ContactId { get; private set; }
        public string EmailAddressId { get; private set; }
        public string Address { get; set; }
        public EmailAddressType Type { get; set; }

        #region Methods

        #region Instace methods

        public void Update()
        {
            try
            {
                DataAccess.Update(this);
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's email address couldn't be updated. " + ex.Message);
            }
        }

        #endregion

        #region Static methods

        //  A contact must not have duplicate Phone Numbers.
        //  A contact can have at most 5 phone numbers.
        internal static EmailAddress Add(string contactId, string address, EmailAddressType type)
        {
            var emails = Add(contactId, new string[1] { address }, type);
            return emails.ElementAt(0);
        }

        //  A contact must not have duplicate email addresses.
        //  A contact can have at most 5 email addresses.
        internal static IEnumerable<EmailAddress> Add(string contactId, IEnumerable<string> addresses, EmailAddressType type)
        {
            //  Consider unique email addresses only
            addresses = addresses.Distinct();

            int newEmailAddressesCount = addresses.Count();
            int actualEmailAddressesCount = CountByContact(contactId);

            if ((actualEmailAddressesCount + newEmailAddressesCount) > 5)
                throw new Exception("A contact can't have more than 5 email addresses.");

            var emails = new List<EmailAddress>();
            foreach (string address in addresses)
            {
                if (Exists(contactId, address))
                    throw new Exception(string.Format("The contact '{0}' already has the phone number '{1}'.", contactId, address));

                emails.Add(CreateNewEmailAddress(contactId, address, type));
            }

            DataAccess.Add(emails);

            return emails;
        }

        internal static void Delete(EmailAddress email)
        {
            DataAccess.Delete(email);
        }

        internal static void DeleteByContact(string contactId)
        {
            DataAccess.DeleteByContact(contactId);
        }

        internal static List<EmailAddress> GetByContact(string contactId)
        {
            return DataAccess.GetByContact(contactId);
        }

        private static bool Exists(string contactId, string address)
        {
            return DataAccess.Exists(contactId, address);
        }

        private static int CountByContact(string contactId)
        {
            return DataAccess.CountByContact(contactId);
        }

        private static EmailAddress CreateNewEmailAddress(string contactId, string address, EmailAddressType type)
        {
            string emailAddressId = Guid.NewGuid().ToString().Replace("-", string.Empty);
            return new EmailAddress(contactId, emailAddressId, address, type);
        }

        #endregion

        #endregion

        #region Internal classes

        //  I create this class as "private" to avoid other classes to manipulate contacts email addresses' data directly 
        private static class DataAccess
        {
            #region Methods

            #region Public 

            public static void Add(EmailAddress email)
            {
                var entity = CreateEntityFromObject(email);
                if (entity == null)
                    return;

                var ctx = new EmailAddressesContext();
                ctx.Add(entity);
            }

            public static void Add(IEnumerable<EmailAddress> emails)
            {
                var entities = new List<EmailAddressEntity>();
                foreach (var email in emails)
                {
                    var entity = CreateEntityFromObject(email);
                    if (entity != null)
                        entities.Add(entity);
                }

                var ctx = new EmailAddressesContext();
                ctx.AddInBatch(entities);
            }

            /// <summary>
            /// Indicates whether a contact already has an email address.
            /// </summary>
            /// <param name="contactId">Contact identifier.</param>
            /// <param name="address">Email address.</param>
            /// <returns>Returns a boolean idicating whether the contact has the specified email address or not.</returns>
            public static bool Exists(string contactId, string address)
            {
                var ctx = new EmailAddressesContext();

                var query =
                    from    entity in ctx.Entities
                    where   entity.PartitionKey == EmailAddressEntity.GeneratePartitionKey(contactId) && entity.Address == address
                    select  entity;

                return ctx.Exists(query);
            }

            /// <summary>
            /// Counts the email addresses a contact has.
            /// </summary>
            /// <param name="contactId">Contact identifier.</param>
            /// <returns>Returns the number of email addresses the contact has.</returns>
            public static int CountByContact(string contactId)
            {
                var ctx = new EmailAddressesContext();

                var query =
                    from    entity in ctx.Entities
                    where   entity.PartitionKey == EmailAddressEntity.GeneratePartitionKey(contactId)
                    select  entity;

                return ctx.Count(query);
            }

            public static void Update(EmailAddress email)
            {
                if (email == null)
                    throw new ArgumentNullException("email");

                string partitionKey = EmailAddressEntity.GeneratePartitionKey(email.ContactId);
                string rowKey = EmailAddressEntity.GenerateRowKey(email.EmailAddressId);

                var ctx = new EmailAddressesContext();

                var entity = ctx.Get(partitionKey, rowKey);
                if (entity == null)
                    throw new Exception(string.Format("The contact '{0}' email address '{1}' to update doesn't exist.", email.ContactId, email.EmailAddressId));

                //  The only property that we'll allow to update are the address and type
                //  So those are the only entity's properties we'll set
                entity.Address = email.Address;
                entity.Type = (int)email.Type;

                ctx.Update(entity);
            }

            public static void Delete(EmailAddress email)
            {
                if (email == null)
                    return;

                string partitionKey = EmailAddressEntity.GeneratePartitionKey(email.ContactId);
                string rowKey = EmailAddressEntity.GenerateRowKey(email.EmailAddressId);

                var ctx = new EmailAddressesContext();
                ctx.Delete(partitionKey, rowKey);
            }

            public static void DeleteByContact(string contactId)
            {
                string partitionKey = EmailAddressEntity.GeneratePartitionKey(contactId);

                var ctx = new EmailAddressesContext();
                var query =
                    from    entity in ctx.Entities
                    where   entity.PartitionKey == partitionKey
                    select  entity;

                ctx.DeleteInBatch(query);
            }

            public static List<EmailAddress> GetByContact(string contactId)
            {
                string partitionKey = EmailAddressEntity.GeneratePartitionKey(contactId);

                var ctx = new EmailAddressesContext();
                var entities = ctx.GetByPartitionKey(partitionKey);
                return MaterializeCollectionFromEntities(entities);
            }

            #endregion

            #region Private 

            private static EmailAddressEntity CreateEntityFromObject(EmailAddress obj)
            {
                if (obj == null)
                    return null;

                return new EmailAddressEntity
                {
                    ContactId = obj.ContactId,
                    EmailAddressId = obj.EmailAddressId,
                    Address = obj.Address,
                    Type = (int)obj.Type
                };
            }

            private static EmailAddress MaterializeFromEntity(EmailAddressEntity entity)
            {
                if (entity == null)
                    return null;

                return new EmailAddress(entity.ContactId, entity.EmailAddressId, entity.Address, (EmailAddressType)entity.Type);
            }

            private static List<EmailAddress> MaterializeCollectionFromEntities(IEnumerable<EmailAddressEntity> entities)
            {
                var list = new List<EmailAddress>();
                foreach (var entity in entities)
                {
                    var obj = MaterializeFromEntity(entity);
                    if (obj != null)
                        list.Add(obj);
                }
                return list;
            }

            #endregion 

            #endregion

            #region Internal classes

            class EmailAddressEntity : TableServiceEntity
            {
                #region Properties

                private string _contactId;
                public string ContactId 
                {
                    get { return _contactId; }
                    set
                    {
                        _contactId = value;
                        PartitionKey = GeneratePartitionKey(value);
                    }
                }

                private string _emailAddressId;
                public string EmailAddressId 
                {
                    get { return _emailAddressId; }
                    set
                    {
                        _emailAddressId = value;
                        RowKey = GenerateRowKey(value);
                    }
                }

                public string Address { get; set; }
                public int Type { get; set; }

                #endregion

                #region Methods

                public static string GeneratePartitionKey(string contactId)
                {
                    //  This will partition email addresses by contact
                    return contactId;
                }

                public static string GenerateRowKey(string emailAddressId)
                {
                    return emailAddressId;
                }

                #endregion
            }

            class EmailAddressesContext : GenericTableContext<EmailAddressEntity>
            {
                /// <summary>
                /// Initializes context using a parameterized Cloud Storage Account.
                /// </summary>
                public EmailAddressesContext()
                    : this(Parameters.Internal.GetCloudStorageAccount())
                { }

                /// <summary>
                /// Initializes context using the specified Cloud Storage Account.
                /// </summary>
                /// <param name="account"></param>
                public EmailAddressesContext(CloudStorageAccount account)
                    : base(account, "ContactsEmailAddresses")
                { }
            }

            #endregion
        }

        #endregion
    }

    public static class EmailAddressListExtensions
    {
        public static EmailAddress FindById(this IEnumerable<EmailAddress> list, string contactId, string emailAddressId)
        {
            return list.FirstOrDefault(e => e.ContactId == contactId && e.EmailAddressId == emailAddressId);
        }
    }
}
