﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Services.Client;

using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;

using mana.Data.Azure.Tables;

namespace Library
{
    public enum PhoneNumberType
    {
        Home = 0,
        PersonalCell = 1,
        Work = 2,
        WorkCell = 3
    }

    [Serializable]
    public class PhoneNumber
    {
        public PhoneNumber(string contactId, string phoneNumberId, string number, PhoneNumberType type)
        {
            ContactId = contactId;
            PhoneNumberId = phoneNumberId;
            Number = number;
            Type = type;
        }

        public string ContactId { get; private set; }
        public string PhoneNumberId { get; private set; }
        public string Number { get; set; }
        public PhoneNumberType Type { get; set; }

        #region Methods

        #region Instance methods

        public void Update()
        {
            try
            {
                DataAccess.Update(this);
            }
            catch (Exception ex)
            {
                throw new Exception("The contact's phone number couldn't be updated. " + ex.Message);
            }
        }

        #endregion

        #region Static methods

        //  A contact must not have duplicate Phone Numbers.
        //  A contact can have at most 5 phone numbers.
        internal static PhoneNumber Add(string contactId, string number, PhoneNumberType type)
        {
            var phones = Add(contactId, new string[1] { number }, type);
            return phones.ElementAt(0);
        }

        //  A contact must not have duplicate Phone Numbers.
        //  A contact can have at most 5 phone numbers.
        internal static IEnumerable<PhoneNumber> Add(string contactId, IEnumerable<string> numbers, PhoneNumberType type)
        {
            //  Consider unique phone numbers only
            numbers = numbers.Distinct();

            int newPhoneNumbersCount = numbers.Count();
            int actualPhoneNumbersCount = CountByContact(contactId);

            if ((actualPhoneNumbersCount + newPhoneNumbersCount) > 5)
                throw new Exception("A contact can't have more than 5 phone numbers.");

            var phones = new List<PhoneNumber>();
            foreach(string number in numbers)
            {
                if (Exists(contactId, number))
                    throw new Exception(string.Format("The contact '{0}' already has the phone number '{1}'.", contactId, number));
                
                phones.Add(CreateNewPhoneNumber(contactId, number, type));
            }

            DataAccess.Add(phones);

            return phones;
        }

        internal static void Delete(PhoneNumber phone)
        {
            DataAccess.Delete(phone);
        }

        internal static void DeleteByContact(string contactId)
        {
            DataAccess.DeleteByContact(contactId);
        }

        internal static List<PhoneNumber> GetByContact(string contactId)
        {
            return DataAccess.GetByContact(contactId);
        }

        private static bool Exists(string contactId, string number)
        {
            return DataAccess.Exists(contactId, number);
        }

        private static int CountByContact(string contactId)
        {
            return DataAccess.CountByContact(contactId);
        }

        private static PhoneNumber CreateNewPhoneNumber(string contactId, string number, PhoneNumberType type)
        {
            string phoneNumberId = Guid.NewGuid().ToString().Replace("-", string.Empty);
            return new PhoneNumber(contactId, phoneNumberId, number, type);
        }

        #endregion

        #endregion

        #region Internal classes

        //  I create this class as "private" to avoid other classes to manipulate contacts phone numbers' data directly 
        private static class DataAccess
        {
            #region Methods

            #region Public 

            public static void Add(PhoneNumber phone)
            {
                var entity = CreateEntityFromObject(phone);
                if (entity == null)
                    return;

                var ctx = new PhoneNumbersContext();
                ctx.Add(entity);
            }

            public static void Add(IEnumerable<PhoneNumber> phones)
            {
                var entities = new List<PhoneNumberEntity>();
                foreach (var phone in phones)
                {
                    var entity = CreateEntityFromObject(phone);
                    if (entity != null)
                        entities.Add(entity);
                }

                var ctx = new PhoneNumbersContext();
                ctx.AddInBatch(entities);
            }

            /// <summary>
            /// Indicates whether a contact already has a phone number.
            /// </summary>
            /// <param name="contactId">Contact identifier.</param>
            /// <param name="number">Phone number.</param>
            /// <returns>Returns a boolean idicating whether the contact has the specified phone number or not.</returns>
            public static bool Exists(string contactId, string number)
            {
                var ctx = new PhoneNumbersContext();

                var query =
                    from    entity in ctx.Entities
                    where   entity.PartitionKey == PhoneNumberEntity.GeneratePartitionKey(contactId) && entity.Number == number
                    select  entity;

                return ctx.Exists(query);
            }

            /// <summary>
            /// Counts the phone numbers a contact has.
            /// </summary>
            /// <param name="contactId">Contact identifier.</param>
            /// <returns>Returns the number of phone numbers the contact has.</returns>
            public static int CountByContact(string contactId)
            {
                var ctx = new PhoneNumbersContext();

                var query =
                    from    entity in ctx.Entities
                    where   entity.PartitionKey == PhoneNumberEntity.GeneratePartitionKey(contactId)
                    select  entity;

                return ctx.Count(query);
            }

            public static void Update(PhoneNumber phone)
            {
                if (phone == null)
                    throw new ArgumentNullException("phone");

                string partitionKey = PhoneNumberEntity.GeneratePartitionKey(phone.ContactId);
                string rowKey = PhoneNumberEntity.GenerateRowKey(phone.PhoneNumberId);

                var ctx = new PhoneNumbersContext();

                var entity = ctx.Get(partitionKey, rowKey);
                if (entity == null)
                    throw new Exception(string.Format("The contact '{0}' phone number '{1}' to update doesn't exist.", phone.ContactId, phone.PhoneNumberId));

                //  The only property that we'll allow to update are the number and type
                //  So those are the only entity's properties we'll set
                entity.Number = phone.Number;
                entity.Type = (int)phone.Type;

                ctx.Update(entity);
            }

            public static void Delete(PhoneNumber phone)
            {
                if (phone == null)
                    throw new ArgumentNullException("phone");

                string partitionKey = PhoneNumberEntity.GeneratePartitionKey(phone.ContactId);
                string rowKey = PhoneNumberEntity.GenerateRowKey(phone.PhoneNumberId);

                var ctx = new PhoneNumbersContext();
                ctx.Delete(partitionKey, rowKey);
            }

            public static void DeleteByContact(string contactId)
            {
                string partitionKey = PhoneNumberEntity.GeneratePartitionKey(contactId);

                var ctx = new PhoneNumbersContext();
                var query =
                    from    entity in ctx.Entities
                    where   entity.PartitionKey == partitionKey
                    select  entity;

                ctx.DeleteInBatch(query);
            }

            public static List<PhoneNumber> GetByContact(string contactId)
            {
                string partitionKey = PhoneNumberEntity.GeneratePartitionKey(contactId);

                var ctx = new PhoneNumbersContext();
                var entities = ctx.GetByPartitionKey(partitionKey);
                return MaterializeCollectionFromEntities(entities);
            }

            #endregion

            #region Private

            private static PhoneNumberEntity CreateEntityFromObject(PhoneNumber obj)
            {
                if (obj == null)
                    return null;

                return new PhoneNumberEntity
                {
                    ContactId = obj.ContactId,
                    PhoneNumberId = obj.PhoneNumberId,
                    Number = obj.Number,
                    Type = (int)obj.Type
                };
            }

            private static PhoneNumber MaterializeFromEntity(PhoneNumberEntity entity)
            {
                if (entity == null)
                    return null;

                return new PhoneNumber(entity.ContactId, entity.PhoneNumberId, entity.Number, (PhoneNumberType)entity.Type);
            }

            private static List<PhoneNumber> MaterializeCollectionFromEntities(IEnumerable<PhoneNumberEntity> entities)
            {
                var list = new List<PhoneNumber>();
                foreach (var entity in entities)
                {
                    var obj = MaterializeFromEntity(entity);
                    if (obj != null)
                        list.Add(obj);
                }
                return list;
            }

            #endregion

            #endregion

            #region Internal classes

            class PhoneNumberEntity : TableServiceEntity
            {
                #region Properties

                private string _contactId;
                public string ContactId
                {
                    get { return _contactId; }
                    set
                    {
                        _contactId = value;
                        PartitionKey = GeneratePartitionKey(value);
                    }
                }

                private string _phoneNumberId;
                public string PhoneNumberId
                {
                    get { return _phoneNumberId; }
                    set
                    {
                        _phoneNumberId = value;
                        RowKey = GenerateRowKey(value);
                    }
                }

                public string Number { get; set; }
                public int Type { get; set; }

                #endregion

                #region Methods

                public static string GeneratePartitionKey(string contactId)
                {
                    //  This will partition phone numbers by contact
                    return contactId;
                }

                public static string GenerateRowKey(string phoneNumberId)
                {
                    return phoneNumberId;
                }

                #endregion
            }

            class PhoneNumbersContext : GenericTableContext<PhoneNumberEntity>
            {
                /// <summary>
                /// Initializes context using a parameterized Cloud Storage Account.
                /// </summary>
                public PhoneNumbersContext()
                    : this(Parameters.Internal.GetCloudStorageAccount())
                { }

                /// <summary>
                /// Initializes context using the specified Cloud Storage Account.
                /// </summary>
                /// <param name="account"></param>
                public PhoneNumbersContext(CloudStorageAccount account)
                    : base(account, "ContactsPhoneNumbers")
                { }
            }

            #endregion
        }

        #endregion
    }

    public static class PhoneNumberListExtensions
    {
        public static PhoneNumber FindById(this IEnumerable<PhoneNumber> list, string contactId, string phoneNumberId)
        {
            return list.FirstOrDefault(pn => pn.ContactId == contactId && pn.PhoneNumberId == phoneNumberId);
        }
    }
}
