﻿/*
    Copyright (c) 2012, Matías López
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted 
    provided that the following conditions are met:
        *   Redistributions of source code must retain the above copyright notice, this list of 
            conditions and the following disclaimer.
        *   Redistributions in binary form must reproduce the above copyright notice, this list of 
            conditions and the following disclaimer in the documentation and/or other materials 
            provided with the distribution.
        *   Neither the name of Matías López nor the names of its contributors may be used to 
            endorse or promote products derived from this software without specific prior written 
            permission.
    
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
    AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
    OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
    POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Collections.ObjectModel;
using System.Collections.Generic;
using System.Text;

using Microsoft.WindowsAzure;

namespace mana.Data.Azure.Tables
{
    public enum TablesModelSynchronizationOption
    {
        PreserveAzureTableIfNotInModel = 0,
        DeleteAzureTableIfNotInModel = 1
    }

    /// <summary>
    /// Represents a complete Model of Azure Tables.
    /// </summary>
    public class TableModel
    {
        private List<string> _tableNames;
        private TableManager _azureTableManager;

        /// <summary>
        /// Initializes an instance of the TableModel with the targeted Cloud Storage Account.
        /// </summary>
        /// <param name="account"></param>
        public TableModel(CloudStorageAccount account)
        {
            _azureTableManager = new TableManager(account);
            _tableNames = new List<string>();
        }

        #region Properties

        /// <summary>
        /// Model's Azure Table names.
        /// </summary>
        public List<string> TableNames
        {
            get { return _tableNames; }
        }

        #endregion Properties

        #region Methods

        /// <summary>
        /// Apply this model to the Storage Account. By default, it preserves the Azure Tables not present in this model on Storage Account.
        /// </summary>
        public void Synchronize()
        {
            Synchronize(TablesModelSynchronizationOption.PreserveAzureTableIfNotInModel);
        }

        /// <summary>
        /// Apply this model to the Storage Account with the specified synchronization option. 
        /// </summary>
        /// <param name="option">Specifies if the Azure Tables on Storage will be preserved or deleted if it does not exist on this model.</param>
        public void Synchronize(TablesModelSynchronizationOption option)
        {
            foreach (string tableName in TableNames)
            {
                _azureTableManager.Add(tableName);
            }

            switch (option)
            {
                case TablesModelSynchronizationOption.PreserveAzureTableIfNotInModel:
                    //  nothing, for now
                    break;
                case TablesModelSynchronizationOption.DeleteAzureTableIfNotInModel:
                    IEnumerable<string> existingTableNames = _azureTableManager.GetNames();
                    foreach (string existingTableName in existingTableNames)
                    {
                        if (!TableNames.Contains(existingTableName))
                            _azureTableManager.Delete(existingTableName);
                    }
                    
                    break;
                default:
                    throw new Exception("Invalid Azure Tables Model synchronization option.");
            }
        }

        #endregion Methods
    }
}
